@@ -1,13 +1,11 @@
 # Net::CIDR
 #
-# Copyright 2001-2003 Sam Varshavchik.
+# Copyright 2001-2010 Sam Varshavchik.
 #
 # with contributions from David Cantrell.
 #
 # This program is free software; you can redistribute it
 # and/or modify it under the same terms as Perl itself.
-#
-# $Revision: 1.14 $
 
 package Net::CIDR;
 
@@ -19,8 +17,6 @@ require Exporter;
 # use AutoLoader qw(AUTOLOAD);
 use Carp;
 
-use Math::BigInt;
-
 @ISA = qw(Exporter);
 
 # Items to export into callers namespace by default. Note: do not export
@@ -54,7 +50,7 @@ use Math::BigInt;
 	
 );
 
-$VERSION = "0.08";
+$VERSION = "0.14";
 
 1;
 
@@ -235,7 +231,7 @@ cidr2range() does not merge adjacent or overlapping netblocks in
 The addr2cidr function takes an IP address and returns a list of all
 the CIDR netblocks it might belong to:
 
-    @a=Net::CIDR::addr2range('192.68.0.31');
+    @a=Net::CIDR::addr2cidr('192.68.0.31');
 
 The result is a thirtythree-element array:
 ('192.68.0.31/32', '192.68.0.30/31', '192.68.0.28/30', '192.68.0.24/29',
@@ -243,7 +239,7 @@ The result is a thirtythree-element array:
 consisting of all the possible subnets containing this address from
 0.0.0.0/0 to address/32.
 
-Any addresses supplied to addr2range after the first will be ignored.
+Any addresses supplied to addr2cidr after the first will be ignored.
 It works similarly for IPv6 addresses, returning a list of one hundred
 and twenty nine elements.
 
@@ -285,15 +281,13 @@ sub cidr2range {
 	my ($ip, $pfix)=($1, $2);
 
 	my $isipv6;
-	($isipv6, $ip)=_ipv6to4($ip);
 
-	my @ips= split (/\.+/, $ip);
+	my @ips=_iptoipa($ip);
 
-	grep {
-	    croak unless $_ >= 0 && $_ <= 255 && $_ =~ /^[0-9]+$/;
-	} @ips;
+	$isipv6=shift @ips;
 
-	croak unless $pfix >= 0 && $pfix <= ($#ips+1) * 8 && $pfix =~ /^[0-9]+$/;
+	croak "$pfix, as in '$cidr', does not make sense"
+	    unless $pfix >= 0 && $pfix <= ($#ips+1) * 8 && $pfix =~ /^[0-9]+$/;
 
 	my @rr=_cidr2iprange($pfix, @ips);
 
@@ -418,6 +412,28 @@ sub _ipv4to6 {
     return join(":", splice (@words, 0, $ind)) . "::"
 	. join(":", @s);
 }
+
+# An IP address to an octet list.
+
+# Returns a list. First element, flag: true if it was an IPv6 flag. Remaining
+# values are octets.
+
+sub _iptoipa {
+    my $iparg=shift;
+
+    my $isipv6;
+    my $ip;
+
+    ($isipv6, $ip)=_ipv6to4($iparg);
+
+    my @ips= split (/\.+/, $ip);
+
+    grep {
+	croak "$_, in $iparg, is not a byte" unless $_ >= 0 && $_ <= 255 && $_ =~ /^[0-9]+$/;
+    } @ips;
+
+    return ($isipv6, @ips);
+}
     
 sub _h62d {
     my $h=shift;
@@ -473,39 +489,45 @@ sub _cidr2iprange {
 #
 
 sub addr2cidr {
-	my(@blocks, $address, $bitmask, $net, $octetstash) = ();
-	my($isIPv6, $addr) = _ipv6to4(shift);
-	my $bitsInAddress = ($isIPv6)?128:32;
+    my @ips=_iptoipa(shift);
 
-	# create BigInts if necessary.  Straight assignment of ints to
-	# BigInts breaks BigInt-ness hence the bizarre subtractions later.
-	if($isIPv6) {
-		$address = new Math::BigInt 0;
-		$bitmask = new Math::BigInt 0;
-		$net = new Math::BigInt 0;
-		$octetstash = new Math::BigInt 0;
-	} else { ($address, $bitmask, $net, $octetstash) = (0, 0, 0, 0); }
+    my $isipv6=shift @ips;
 
-	# convert dotted-octets into an int (or BigInt)
-	do { $address = $address << 8; $address = $address + $_ }
-		for split(/\./, $addr);
+    my $nbits;
 
-	foreach my $bits (reverse 0..$bitsInAddress) {
-		$octetstash = $octetstash - $octetstash + 255;
-		$bitmask = $bitmask - $bitmask;
-		for(1 .. $bits) { $bitmask = ($bitmask << 1) + 1; }
-		$bitmask = $bitmask << ($bitsInAddress - $bits);
-		
-		$net = $address & $bitmask;
-		push @blocks, (map { $_ =~ s/\+//; (($isIPv6)?_ipv4to6($_):$_) }
-		    join('.', reverse map {
-			my $temp = ($octetstash & $net) >> ($_ * 8);
-			$octetstash = $octetstash << 8;
-			$temp;
-		    } (0 .. $bitsInAddress / 8 - 1)
-		))[0]."/$bits";
+    if ($isipv6)
+    {
+	croak "An IPv6 address is 16 bytes long" unless $#ips == 15;
+	$nbits=128;
+    }
+    else
+    {
+	croak "An IPv4 address is 4 bytes long" unless $#ips == 3;
+	$nbits=32;
+    }
+
+    my @blocks;
+
+    foreach my $bits (reverse 0..$nbits)
+    {
+	my @ipcpy=@ips;
+
+	my $n=$bits;
+
+	while ($n < $nbits)
+	{
+	    @ipcpy[$n / 8] &= (0xFF00 >> ($n % 8));
+
+	    $n += 8;
+
+	    $n &= 0xF8;
 	}
-	return @blocks;
+
+	my $s=join(".", @ipcpy);
+
+	push @blocks, ($isipv6 ? _ipv4to6($s):$s) . "/$bits";
+    }
+    return @blocks;
 }
 
 # Address and netmask to CIDR
@@ -515,7 +537,7 @@ sub addrandmask2cidr {
 	my($a_isIPv6) = _ipv6to4($address);
         my($n_isIPv6, $netmask) = _ipv6to4(shift);
 	die("Both address and netmask must be the same type")
-	    if($a_isIPv6 != $n_isIPv6);
+	    if( defined($a_isIPv6) && defined($n_isIPv6) && $a_isIPv6 != $n_isIPv6);
         my $bitsInNetmask = 0;
         my $previousNMoctet = 255;
         foreach my $octet (split/\./, $netmask) {
@@ -527,7 +549,7 @@ sub addrandmask2cidr {
 		($octet == 252) ? 6 :
 		($octet == 248) ? 5 :
 		($octet == 240) ? 4 :
-		($octet == 232) ? 3 :
+		($octet == 224) ? 3 :
 		($octet == 192) ? 2 :
 		($octet == 128) ? 1 :
 		($octet == 0) ? 0 :
@@ -843,15 +865,12 @@ sub cidr2octets {
 
 	my $isipv6;
 
-	($isipv6, $ip)=_ipv6to4($ip);
+	my @ips=_iptoipa($ip);
 
-	my @ips= split (/\.+/, $ip);
+	$isipv6=shift @ips;
 
-	grep {
-	    croak unless $_ >= 0 && $_ <= 255 && $_ =~ /^[0-9]+$/;
-	} @ips;
-
-	croak unless $pfix >= 0 && $pfix <= ($#ips+1) * 8 && $pfix =~ /^[0-9]+$/;
+	croak "$pfix, as in '$cidr', does not make sense"
+	    unless $pfix >= 0 && $pfix <= ($#ips+1) * 8 && $pfix =~ /^[0-9]+$/;
 
 	my $i;
 
@@ -866,7 +885,23 @@ sub cidr2octets {
 
 	if ($#ips < 0 || $bitsleft == 0)
 	{
-	    if ($isipv6)
+	    if ($pfix == 0 && $bitsleft == 0)
+	    {
+		foreach (0..255)
+		{
+		    my @n=($_);
+
+		    if ($isipv6)
+		    {
+			_push_ipv6_octets(\@r, \@n);
+		    }
+		    else
+		    {
+			push @r, $n[0];
+		    }
+		}
+	    }
+	    elsif ($isipv6)
 	    {
 		_push_ipv6_octets(\@r, \@msb);
 	    }
@@ -1001,7 +1036,7 @@ sub cidradd {
 	$b[$i]=undef;
     }
 
-    unless (_ipcmp($lo, $hi) > 0)
+    unless ((! defined $lo) || (! defined $hi) || _ipcmp($lo, $hi) > 0)
     {
 	push @a, $lo;
 	push @b, $hi;
@@ -1154,10 +1189,16 @@ sub cidrlookup {
 
 =head2 $ip=Net::CIDR::cidrvalidate($ip);
 
-Validate whether $ip is a valid IPv4 or IPv6 address.
+Validate whether $ip is a valid IPv4 or IPv6 address, or a CIDR.
 Returns its argument or undef.
 Spaces are removed, and IPv6 hexadecimal address are converted to lowercase.
 
+If $ip contains a "/", it must be a valid CIDR, otherwise it must be a valid
+IPv4 or an IPv6 address.
+
+A technically invalid CIDR, such as "192.168.0.1/24" fails validation, returning
+undef.
+
 =cut
 
 sub cidrvalidate {
@@ -1167,6 +1208,16 @@ sub cidrvalidate {
 
     $v=lc($v);
 
+    my $suffix;
+
+    ($v, $suffix)=($1, $2) if $v =~ m@(.*)/(.*)@;
+
+    if (defined $suffix)
+    {
+	return undef unless $suffix =~ /^\d+$/ &&
+	    ($suffix eq "0" || $suffix =~ /^[123456789]/);
+    }
+
     if ($v =~ /^([0-9\.]+)$/ || $v =~ /^::ffff:([0-9\.]+)$/ ||
 	$v =~ /^:([0-9\.]+)$/)
     {
@@ -1183,7 +1234,25 @@ sub cidrvalidate {
 	    return if /^0./;
 	    return if $_ < 0 || $_ > 255;
 	}
-	return $v;
+
+	if ($v =~ /^::ffff/)
+	{
+	    $suffix=128 unless defined $suffix;
+
+	    return undef if $suffix < 128-32;
+
+	    $suffix -= 128-32;
+	}
+	else
+	{
+	    $suffix=32 unless defined $suffix;
+	}
+
+	foreach (addr2cidr($n))
+	{
+	    return $v if $_ eq "$n/$suffix";
+	}
+	return undef;
     }
 
     return undef unless $v =~ /^[0-9a-f:]+$/;
@@ -1193,14 +1262,20 @@ sub cidrvalidate {
 
     my @o=grep (/./, split(/:/, $v));
 
-    return undef if $#o >= 8;
+    return undef if ($#o >= 8 || ($#o<7 && $v !~ /::/));
 
     foreach (@o)
     {
 	return undef if length ($_) > 4;
     }
 
-    return $v;
+    $suffix=128 unless defined $suffix;
+
+    foreach (addr2cidr($v))
+    {
+	return $v if $_ eq "$v/$suffix";
+    }
+    return undef;
 }
 
 =pod
@@ -1208,7 +1283,7 @@ sub cidrvalidate {
 =head1 BUGS
 
 Garbage in, garbage out.
-Always use validate() before doing anything with untrusted input.
+Always use cidrvalidate() before doing anything with untrusted input.
 Otherwise,
 "slightly" invalid input will work (extraneous whitespace
 is generally OK),
